// Using LittleFS/SPIFFS. Using a file(ascii+nonascii) to load/store variables, data, parameters. Not thread safe. by ken woo.
//
// * version 0.0. 20210319. initial version.
//
// * version 0.1. 20210320.
// change some names and reorder others for consistency.
// add delete record function.
// add renew function. because the deleted record ones still exist in file. renew is clone a clean one and remove old.
// add list records names and ids.
//
// # NOTE: delete/list records and renew methods are only viable when file is at the Closed() state;
// since fopen mode is different while applied.
//
// # NOTE: As for the replace method is concerned, it could be done by delete then add.


#include <ESP8266WiFi.h>
#include<LittleFS.h>


#define FILESYSTEM LittleFS
//#define FILESYSTEM SPIFFS // its format() takes too long.


typedef class cDataBase{

    #define PREAMBLE        (0x55A55AAA)
    #define DEPREAMBLE      (0x55A55AAB)
    // a complete record unit in a file:
    // [4-byte preamble]+[4-byte name length]+[4-byte name checksum]+[4-byte record length]+[n-byte name]+[m-byte record contents]=16+m+n.
    // payload is 16kbytes/1000records.

    typedef struct __attribute__((__packed__)) sRecord{
        unsigned nlen; // name length.
        unsigned ncs; // name checksum.
        unsigned rlen; // record length, name+id+data.
    } sRecord;

    static void setup_filesystem(){
        if (!filesystem_opened){ // load FS whenever construct, unload whenever destruct. counter used.
            FILESYSTEM.begin();
        }
        filesystem_opened++;

        static int x=1;
        if (x){
            x=0;
            {
                Dir dir = FILESYSTEM.openDir("/");
                if (!dir.next()){
                    FILESYSTEM.format();
                    printf("\r\n\r\nfilesystem formatted.\r\n");
                }
            }
            FSInfo fs_info;
            if (FILESYSTEM.info(fs_info)){
                printf(\
                    "--------\r\nThe FileSystem:\r\nentire space %d bytes\r\n%d bytes used\r\n"\
                    "free space %d bytes\r\nblock size %d bytes\r\npage size %d blocks\r\n--------\r\n",\
                    fs_info.totalBytes, fs_info.usedBytes, fs_info.totalBytes-fs_info.usedBytes, fs_info.blockSize, fs_info.pageSize);
            }
            Dir dir = FILESYSTEM.openDir("/");
            printf("list root directory \"/\":\r\n");
            while (dir.next())
                printf("%s    %s\r\n", dir.fileName().c_str(), dir.isDirectory()? "<DIR>": (String(dir.fileSize())+" bytes").c_str());
        }
    };

    unsigned my_find(const void *str, unsigned sz){ // only for either char or unsigned char. 0/not found. others/the found position.
        unsigned i=0, j=0, w=sz*2, sum=0, csum=0, start_p=0, end_p=sz;
        unsigned char u[w+sz];
        int v=0;

        if (fh.readBytes((char*)u, sz)!=sz) return 0;
        for (; i<sz; i++){
            sum+=((const unsigned char*)str)[i];
            csum+=u[i];
        }

        while (1){
            if (sum==csum){
                for (i=0, j=0; i<sz && (((const unsigned char*)str)[i]==u[start_p+i-j]); i++) if (start_p+i-j==sz+w-1) j=start_p+i+1;
                if (i==sz){
                    fh.seek(-v, SeekCur);
                    return fh.position();
                }
            }
            if (!v){
                if (end_p==sz) v=fh.readBytes((char*)(u+end_p), w);
                else if (end_p==w+sz) v=fh.readBytes((char*)u, w);
                else {v=fh.readBytes((char*)(u+w), sz); v+=fh.readBytes((char*)u, sz);}
                if (!v) return v;
            }
            if (start_p==w+sz) start_p=0;
            if (end_p==w+sz) end_p=0;
            csum+=u[end_p++];
            csum-=u[start_p++];
            v--;
        }
    };

    static unsigned filesystem_opened;
    const unsigned preamble=PREAMBLE;
    const unsigned depreamble=DEPREAMBLE;

    const char *db_name;
    const char *fmode;
    File fh;

    public:

    cDataBase(): db_name(0), fmode(0){setup_filesystem();};
    cDataBase(const char *dbn): db_name(dbn), fmode(0){setup_filesystem();};
    void operator=(const char *dbn){
        this->Close();
        db_name=dbn;
    };
    ~cDataBase(){
        this->Close();
        if (!--filesystem_opened) FILESYSTEM.end();
    };

    void Open(const char *m="a+"){
        if (!db_name || fmode) return;
        fh=FILESYSTEM.open(db_name, m);
        if (fh) fmode=m;
    };

    void Close(){
        if (fmode){
            fh.close();
            fmode=0;
        }
    };

    void DumpAll(int f=0){ // 0: ascii; 1: hex.
        char u[8];
        int v;
        fh.seek(0, SeekSet);
        while ((v=fh.readBytes(u, 8))==8){
            printf("[");
            for (int i=0; i<8; i++)
                if (f) printf("%02X ", u[i]);
                else {
                    if (u[i]<' ' || u[i]>'~') u[i]='.';
                    printf("%2c ", u[i]);
                }
            printf("%s\r\n", "]");
        }
        printf("[");
        for (int i=0; i<v; i++)
            if (f) printf("%02X ", u[i]);
            else {
                if (u[i]<' ' || u[i]>'~') u[i]='.';
                printf("%2c ", u[i]);
            }
        printf("%s\r\n", "]");
    };

    template<typename T>
    void WriteWords(const char *name, int id, T data){
        // for elementary types, id whether unique or not maintained by user.
        // name is for the db to search purpose; whether unique or not maintained by user.

        sRecord u;

        for (u.nlen=0, u.ncs=0; name[u.nlen]; u.ncs+=name[u.nlen], u.nlen++);
        u.rlen=u.nlen+sizeof(id)+sizeof(data);

        fh.write((const unsigned char*)&preamble, sizeof(preamble));
        fh.write((const unsigned char*)&u, sizeof(u));
        fh.write((const unsigned char*)name, u.nlen);
        fh.write((unsigned char*)&id, sizeof(id));
        fh.write((unsigned char*)&data, sizeof(data));
    };

    template<typename T>
    void WriteBytes(const char *name, int id, const T *buffer, unsigned length){
        sRecord u;

        for (u.nlen=0, u.ncs=0; name[u.nlen]; u.ncs+=name[u.nlen], u.nlen++);
        u.rlen=u.nlen+sizeof(id)+sizeof(T)*length; // note, not sizeof(buffer).

        fh.write((const unsigned char*)&preamble, sizeof(preamble));
        fh.write((const unsigned char*)&u, sizeof(u));
        fh.write((const unsigned char*)name, u.nlen);
        fh.write((unsigned char*)&id, sizeof(id));
        fh.write((const unsigned char*)buffer, sizeof(T)*length);
    };

    template<typename T>
    void WriteBytesCont(const T *buffer, unsigned length){
        fh.write((const unsigned char*)buffer, sizeof(T)*length);
    };

    void FindStart(){ // a brand new start to find; from file ptr position 0.
        fh.flush();
        fh.seek(0, SeekSet);
    };

    unsigned FindNext(const char *name, int id=-1, unsigned proc=0){
        // find start from the current file position. file pos is moving up with FindNext.
        // return the fpos points to the begin of data(id+data contents) when id==-1. (and the fpos is where the current fptr pointing to).
        // return the fpos points to the begin of data contents when id!=-1. (and the fpos is where the current fptr pointing to).
        //
        // later, the var proc added:
        // 0/default, 1/delete when name(or and id) matched.
        // special!!! 2/for list name and id. because lacks target can stored the info, the returned var is the name length.
        // id though, is 4-byte but may not proper when directly specifies 4 but do.
        // 3/for renew. again, move fpos to preamble head and return the entire record length including preamble.

        unsigned sum=0, i=0, len=0, pream=0;
        sRecord u;
        for (; name[len]; len++) sum+=name[len];
        while (pream=my_find(&preamble, sizeof(preamble))){
            if (fh.readBytes((char*)&u, sizeof(u))==sizeof(u)){
                if ((len==u.nlen) && (sum==u.ncs)){
                    char n[u.nlen];
                    if (fh.readBytes(n, u.nlen)==u.nlen){
                        for (i=0; i<u.nlen && (name[i]==n[i]); i++);
                        if (i==u.nlen){
                            if (id!=-1){
                                unsigned rid;
                                if (fh.readBytes((char*)&rid, sizeof(rid))==sizeof(rid)){
                                    if (id!=rid) fh.seek(u.rlen-u.nlen-sizeof(rid), SeekCur);
                                    else {
                                        if (proc==1){ // delete. move pos to the preamble head.
                                            fh.seek(pream-sizeof(preamble), SeekSet);
                                            fh.write((const unsigned char*)&depreamble, sizeof(depreamble));
                                            return fh.position();
                                        }
                                        else return fh.position(); // found data
                                    }
                                }
                                else {/*fh.seek(0, SeekEnd); */return 0;}
                            }
                            else {
                                if (proc==1){ // delete. move pos to the preamble head.
                                    fh.seek(pream-sizeof(preamble), SeekSet);
                                    fh.write((const unsigned char*)&depreamble, sizeof(depreamble));
                                    return fh.position();
                                }
                                else return fh.position(); // found id+data
                            }
                        }
                        else fh.seek(u.rlen-u.nlen, SeekCur);
                    }
                    else {/*fh.seek(0, SeekEnd); */return 0;}
                }
                else if ((proc==2) && (id==-1)){ // list name and id
                    return u.nlen;
                }
                else if ((proc==3) && (id==-1)){ // renew
                    fh.seek(pream-sizeof(preamble), SeekSet);
                    return sizeof(preamble)+sizeof(sRecord)+u.rlen;
                }
                else fh.seek(u.rlen, SeekCur);
            }
            else {/*fh.seek(0, SeekEnd); */return 0;}
        }
        /*fh.seek(0, SeekEnd);*/
        return 0;
    };

    template<typename T>
    bool ReadWords(T &data, int &id, const char *name){ // read both the data and id according to name.
        FindStart();
        if (FindNext(name)){
            if (fh.readBytes((char*)&id, sizeof(id))==sizeof(id)){
                if (fh.readBytes((char*)&data, sizeof(data))==sizeof(data)) return true;
                return false;
            }
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadWords(T &data, const char *name, int id=-1){ // read the data only according to name or more with the id when id!=-1.
        FindStart();
        if (FindNext(name, id)){
            if (id==-1) fh.seek(sizeof(id), SeekCur);
            if (fh.readBytes((char*)&data, sizeof(data))==sizeof(data)) return true;
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadWordsNext(T &data, int &id, const char *name){ // read both the data and id according to name. Next means start from current pos.
        if (FindNext(name)){
            if (fh.readBytes((char*)&id, sizeof(id))==sizeof(id)){
                if (fh.readBytes((char*)&data, sizeof(data))==sizeof(data)) return true;
                return false;
            }
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadWordsNext(T &data, const char *name, int id=-1){ // read the data only according to name or more with the id when id!=-1.
        if (FindNext(name, id)){
            if (id==-1) fh.seek(sizeof(id), SeekCur);
            if (fh.readBytes((char*)&data, sizeof(data))==sizeof(data)) return true;
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadBytes(T *buffer, unsigned length, int &id, const char *name){ // read both the data and id according to name.
        FindStart();
        if (FindNext(name)){
            if (fh.readBytes((char*)&id, sizeof(id))==sizeof(id)){
                if (fh.readBytes((char*)buffer, sizeof(T)*length)==sizeof(T)*length) return true; // note, not sizeof(buffer).
                return false;
            }
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadBytes(T *buffer, unsigned length, const char *name, int id=-1){
    // read the data only according to name or more with the id when id!=-1.

        FindStart();
        if (FindNext(name, id)){
            if (id==-1) fh.seek(sizeof(id), SeekCur);
            if (fh.readBytes((char*)buffer, sizeof(T)*length)==sizeof(T)*length) return true; // note, not sizeof(buffer).
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadBytesNext(T *buffer, unsigned length, int &id, const char *name){ // read both the data and id according to name.
        if (FindNext(name)){
            if (fh.readBytes((char*)&id, sizeof(id))==sizeof(id)){
                if (fh.readBytes((char*)buffer, sizeof(T)*length)==sizeof(T)*length) return true; // note, not sizeof(buffer).
                return false;
            }
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadBytesNext(T *buffer, unsigned length, const char *name, int id=-1){
    // read the data only according to name or more with the id when id!=-1.

        if (FindNext(name, id)){
            if (id==-1) fh.seek(sizeof(id), SeekCur);
            if (fh.readBytes((char*)buffer, sizeof(T)*length)==sizeof(T)*length) return true; // note, not sizeof(buffer).
            return false;
        }
        return false;
    };

    template<typename T>
    void ReadBytesCont(T *buffer, unsigned length){
        fh.readBytes((char*)buffer, sizeof(T)*length);
    };

    void LsRecords(){
        if (!db_name || fmode) return;
        this->Open("r+");
        FindStart();
        unsigned len;
        printf("%s contents:\r\n--------\r\n", db_name);
        while (len=FindNext("", -1, 2)){
            char r[len+1];
            unsigned i;
            r[len]=0;
            fh.readBytes(r, len);
            fh.readBytes((char*)&i, 4);
            printf("[%s    %d]\r\n", r, i);
        }
        printf("--------\r\n");
        this->Close();
    };

    bool DeleteRecords(const char *name, int id=-1){ // delete the data according to the name(or and id), all of which are deleted.
        if (!db_name || fmode) return false;
        this->Open("r+");
        FindStart();
        bool r=FindNext(name, id, 1);
        while (r && FindNext(name, id, 1));
        this->Close();
        return r;
    };

   bool Renew(){
        if (!db_name || fmode) return false;
        int i;
        for (i=0; i<20 && FILESYSTEM.exists(String("/tmp")+i); i++);
        if (i==20) return false;
        String newfile="/tmp"+i;
        File tmp=FILESYSTEM.open(newfile, "w");
        if (!tmp) return false;
        this->Open("r");
        FindStart();
        unsigned r=FindNext("", -1, 3);
        if (!r){
            this->Close();
            tmp.close();
            return false;
        }
        do { //// no care about full.
            {
                char s[r];
                fh.readBytes(s, r);
                tmp.write((unsigned char*)s, r);
            }
        }
        while (r=FindNext("", -1, 3));
        this->Close();
        tmp.close();
        FILESYSTEM.remove(db_name);
        FILESYSTEM.rename(newfile.c_str(), db_name);
        return true;
    };

    static void Ls(const char *path="/"){
        if (!filesystem_opened) FILESYSTEM.begin();

        Dir dir = FILESYSTEM.openDir(path);
        printf("list directory \"%s\":\r\n--------\r\n", path);
        while (dir.next())
            printf("%s    %s\r\n", dir.fileName().c_str(), dir.isDirectory()? "<DIR>": (String(dir.fileSize())+" bytes").c_str());

        FSInfo fs_info;
        if (FILESYSTEM.info(fs_info)){
            printf("\r\nentire space %d bytes\r\n%d bytes used\r\nfree space %d bytes\r\n--------\r\n",\
                fs_info.totalBytes, fs_info.usedBytes, fs_info.totalBytes-fs_info.usedBytes);
        }

        if (!filesystem_opened) FILESYSTEM.end();
    };

} cDataBase;

unsigned cDataBase::filesystem_opened=0;




void setup(){
    Serial.begin(115200);
    delay(5000);

    double THE_PI=3.1415926f;
    cDataBase s;
    s="/database";
    s.Open();

    s.WriteWords("abcd", 4567, 1111.2f);
    s.WriteWords("defg", 7654, 6666);
    s.WriteBytes("1234", 0, "abcdefghijklm", 13);
    s.FindStart();
    printf("[%d %d %d]\r\n", s.FindNext("abcd"), s.FindNext("defg"), s.FindNext("1234"));
    s.DumpAll(1);

    s.WriteWords("doubleTHE_PI", 0, THE_PI);
    s.WriteWords("abcd", 4567, 1111.2f);
    s.WriteWords("fg", 7654, 6666);
    s.WriteBytes("dcba", 2, "abcdefghijklm", 13);
    s.WriteWords("dcba", 4567, 100000.f*100000.f*100000.f);
    s.WriteWords("gd", 1111, 6666);
    s.WriteBytes("eeeeee", 2, "abcdghijklm", 11);

    float L1, L4;
    int I1, I2, I3, I4, I5, I6;
    char C3[14]={0}, C6[12]={0};

    if (s.ReadWords(L4, "dcba", 4567)) printf("dcba[L4 %f]\r\n", L4);
    if (s.ReadBytesNext(C3, 13, I3, "dcba")) printf("dcba[I3a %d, %s]\r\n", I3, C3);

    L4=I4=I3=0;
    if (s.ReadBytes(C3, 13, I3, "dcba")) printf("dcba[I3b %d, %s]\r\n", I3, C3);
    if (s.ReadWordsNext(L4, I4, "dcba")) printf("dcba[L4 %d %f]\r\n", I4, L4);
    if (s.ReadBytesNext(C6, 11, "eeeeee", 1)) printf("eeeeee[C6a %s]\r\n", C6);
    if (s.ReadBytes(C6, 11, "eeeeee", 2)) printf("eeeeee[C6b %s]\r\n", C6);

    if (s.ReadWords(L1, I1, "dcba")) printf("abcd1[%d, %f]\r\n", I1, L1);
    if (s.ReadWords(L1, I1, "abcd")) printf("abcd2[%d, %f]\r\n", I1, L1);

    s.DumpAll(0);

    THE_PI=0;
    if (s.ReadWords(THE_PI, "doubleTHE_PI")) printf("\r\nvariable THE_PI[%f]\r\n", THE_PI);

    s.Close();
    s.LsRecords();
    cDataBase::Ls();

    printf("%d\r\n", s.DeleteRecords("doubleTHE_PI"));s.LsRecords();
    printf("%d\r\n", s.DeleteRecords("eeeeee", 1));s.LsRecords();
    printf("%d\r\n", s.DeleteRecords("abcd"));s.LsRecords();
    printf("%d\r\n", s.DeleteRecords("gd", 1111));s.LsRecords();

    s.Renew();
    s.LsRecords();
    s.Open();
    s.DumpAll(1);
    s.Close();
    cDataBase::Ls();
    s.LsRecords();
}

void loop(){
}
